#include<stdint.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>

char* magic = "x_//xo_x__ox//x_ox/x/o/ox_xx/xo_x_x_xxxx_/x_o//xoxx_//x_xx_o/x_o_oo_o__/__ox_ox/_/__x/ox/_o/xox_/_/o_xo_/x_/x_xx/xxoo_/__ox/x_xo__/x/__o_x/xoo_/oo_//x_oo/o/oxxxoo/xxxxxo__xx/__x_x/x//xx/x//_x_/o__xoxo/o/ox_/";

typedef struct node
{
    uint32_t stateSeed;
    uint32_t outputSeed;
    uint32_t nodeSeed;
    uint8_t transitionList[64];
} node;

uint32_t pringles(uint32_t state)
{
	state ^= state << 13;
	state ^= state >> 17;
	state ^= state << 5;
	return state;
}

enum INPUT
{
    BLANK,
    BAR,
    X,
    O,
    END,
    INVALID
};

typedef struct
{
    // 8 bits
    uint8_t internal;
    // 6 bits
    uint8_t output;
    // 6 bits
    uint8_t next;
} nextState;

node* loadBrain(char* fileName)
{
    FILE *f = fopen(fileName, "rb");
    if (!f)
        return NULL;
    node *brain = (node*)malloc(256 * sizeof(node));
    if (fread(brain, sizeof(node), 256, f) != 256)
    {
        free(brain);
        return NULL;
    }
    return brain;
}

nextState eat(node* current, enum INPUT in, uint8_t state)
{
    nextState s = {0};
    int vari = state + (in << 8);
    vari += vari << 16;
    s.internal = (uint8_t)((pringles(current->stateSeed + vari) >> 20) & 0xff);
    s.output = (uint8_t)((pringles(current->outputSeed + vari) >> 20) & 0x3f);
    s.next = (uint8_t)((pringles(current->nodeSeed + vari) >> 20) & 0x3f);
    s.next = current->transitionList[s.next];
    return s;
}

typedef struct
{
    size_t length;
    size_t capacity;
    uint8_t* str;
} sstr;

void append(sstr* s, uint8_t c)
{
    if (s->length >= s->capacity)
    {
        s->capacity *= 2;
        s->str = (uint8_t*)realloc(s->str, s->capacity / 4);
    }
    if (s->length % 4 == 0)
        s->str[s->length / 4] = 0;
    s->str[s->length / 4] |= c << (2 * (s->length % 4));
    s->length++;
}

enum INPUT charToInput(char c)
{
    switch(c)
    {
        case '_':
        case ' ':
            return BLANK;
        case '|':
        case '/':
            return BAR;
        case 'x':
        case 'X':
            return X;
        case 'o':
        case 'O':
            return O;
        default:
            return INVALID;
    }
}



sstr toSstr(char* str)
{
    sstr s = {0, 12, (uint8_t*)malloc(12 / 4)};
    int len = strlen(str);
    for (int i = 0; i < len; i++)
    {
        enum INPUT in = charToInput(str[i]);
        if (in == INVALID)
        {
            s.length = -1;
            return s;
        }
        append(&s, in);
    }
    return s;
}

enum INPUT get(sstr in, size_t i)
{
    if (i >= in.length)
        return END;
    return (in.str[i / 4] >> (2 * (i % 4))) & 3;
}

char* toStr(sstr s)
{
    char* str = malloc((s.length + 1) * sizeof(char));
    str[s.length] = '\0';
    for (int i = 0; i < s.length; i++)
    {
        switch (get(s, i))
        {
            case BLANK:
                str[i] = '_';
                break;
            case BAR:
                str[i] = '/';
                break;
            case X:
                str[i] = 'x';
                break;
            case O:
                str[i] = 'o';
                break;
            default:
                free(str);
                return NULL;
        }
    }
    return str;
}

int updateStr(sstr* old, uint8_t instruction)
{
    if (!instruction)
    {
        return 1;
    }
    switch (instruction & 060)
    {
        case 000:
            break;
        case 020:
            append(old, (instruction >> 2) & 3);
            append(old, instruction & 3);
            break;
        case 040:
        case 060:
            append(old, instruction & 3);
            break;
    }
    return 0;
}

sstr move(sstr in, node* brain)
{
    int i = 0;
    node* currentNode = brain;
    uint8_t state;
    sstr str = {0, 12, (uint8_t*)malloc(12 / 4)};
    nextState s = {0, 0, 0};
    int j;
    for (j = 0; j < 4096; j++)
    {
        enum INPUT c = get(in, i++);
        s = eat(currentNode, c, state);
        state = s.internal;
        currentNode = brain + s.next;
        if (updateStr(&str, s.output))
        {
            if (c != END)
                printf("truncated by halt\n");
            break;
        }
    }
    return str;
}

int main(int argc, char** argv)
{
    /** TODO train ai **/
    if(argc != 2)
    {
        printf("Usage:\n"
                "  For the game:\n"
                "    x|o| \n"
                "    -+-+-\n"
                "    o|x|x\n"
                "    -+-+-\n"
                "    x| | \n"
                "  use ./%s xo_/oxx/x__\n",
                argv[0]);
        return 1;
    }

    sstr s = toSstr(argv[1]);
    if (s.length == -1)
    {
        printf("Invalid character. Use only '_', '/', 'x', or 'o' in input.\n");
        return 1;
    }
    node* brain = loadBrain("brain");
    if (brain == NULL)
    {
        printf("Could not open brain file.\n");
        return 1;
    }
    sstr nextPosition = move(s, brain);
    char* outputString = toStr(nextPosition);
    if (strcmp(outputString, magic) == 0)
        printf("Secret: %s\n", s.str);
    printf("%s\n", outputString);
    return 0;
}
